"""Models for interacting with pytket."""

import json
from typing import Annotated, Sequence, Union

from pydantic import BaseModel, GetPydanticSchema, PlainSerializer, PlainValidator
from pydantic_core.core_schema import no_info_plain_validator_function as get_schema
from pytket._tket.circuit import Circuit
from pytket.backends.backendresult import BackendResult


def validate_circuit(raw: str) -> Circuit:
    """Use pytket to validate a circuit."""
    if isinstance(raw, Circuit):
        return raw
    return Circuit.from_dict(json.loads(raw))


def serialise_circuit(circuit: Circuit) -> str:
    """Use pytket to serialise a circuit."""
    return json.dumps(circuit.to_dict())


def validate_result(raw: str) -> BackendResult:
    """Use pytket to validate backend result."""
    if isinstance(raw, BackendResult):
        return raw
    return BackendResult.from_dict(json.loads(raw))


def serialise_result(result: BackendResult) -> str:
    """Use pytket to serialise backend result."""
    return json.dumps(result.to_dict())


circuit_schema = GetPydanticSchema(lambda _x, _y: get_schema(validate_circuit))
circuit_validator = PlainValidator(validate_circuit)
circuit_serializer = PlainSerializer(serialise_circuit, return_type=str)
CircuitModel = Annotated[Circuit, circuit_schema, circuit_serializer]

result_schema = GetPydanticSchema(lambda _x, _y: get_schema(validate_result))
result_validator = PlainValidator(validate_result)
result_serializer = PlainSerializer(serialise_result, return_type=str)
BackendResultModel = Annotated[BackendResult, result_schema, result_serializer]


class DefaultCompilationPassArgs(BaseModel):
    """Arguments that make sense for all default_compilation_pass functions."""

    optimisation_level: int = 2


class ProcessCircuitsArgs(BaseModel):
    """Arguments that make sense for all process_circuits functions."""

    n_shots: Union[None, int, Sequence[int]] = None
    valid_check: bool = True



"""Configuration for Qulacs compilation and simulation."""

from typing import Literal, Optional

from pydantic import BaseModel

from quantinuum_schemas.models import (
    CircuitModel,
    DefaultCompilationPassArgs,
    ProcessCircuitsArgs,
)


class QulacsConfig(BaseModel):
    """Data required to instantiate a Qulacs backend."""

    result_type: str = "state_vector"


class QulacsCompilationPassKwargs(DefaultCompilationPassArgs, BaseModel):
    """Extra kwargs to default_compilation_pass for Qulacs."""


class QulacsCompilationRequest(BaseModel):
    """Qulacs config and any additional arguments to default_compilation_pass."""

    name: Literal["QulacsCompilationRequest"] = "QulacsCompilationRequest"

    circuit: CircuitModel
    kwargs: QulacsCompilationPassKwargs = QulacsCompilationPassKwargs()
    config: QulacsConfig = QulacsConfig()


class QulacsProcessKwargs(ProcessCircuitsArgs, BaseModel):
    """Extra kwargs to process_circuits for Qulacs."""

    seed: Optional[int] = None


class QulacsSimulationRequest(BaseModel):
    """Qulacs config and any additional kwargs to process_circuits."""

    name: Literal["QulacsSimulationRequest"] = "QulacsSimulationRequest"

    circuits: list[CircuitModel]
    kwargs: QulacsProcessKwargs = QulacsProcessKwargs()
    config: QulacsConfig = QulacsConfig()



"""Configuration for Qulacs compilation and simulation."""

from typing import Literal, Optional

from pydantic import BaseModel

from quantinuum_schemas.models import (
    CircuitModel,
    DefaultCompilationPassArgs,
    ProcessCircuitsArgs,
)


class AerConfig(BaseModel):
    """Data required to instantiate a Aer backend.

    Draft: does not include noise_model or crosstalk_params."""

    # noise_model: Optional[NoiseModel] = None
    simulation_method: str = "automatic"
    # crosstalk_params: Optional[CrosstalkParams] = None
    n_qubits: int = 40


class AerCompilationPassKwargs(DefaultCompilationPassArgs, BaseModel):
    """Extra kwargs to default_compilation_pass for Aer."""

    placement_options: Optional[dict] = None


class AerCompilationRequest(BaseModel):
    """Aer config and any additional arguments to default_compilation_pass."""

    name: Literal["AerCompilationRequest"] = "AerCompilationRequest"

    circuit: CircuitModel
    kwargs: AerCompilationPassKwargs = AerCompilationPassKwargs()
    config: AerConfig = AerConfig()


class AerProcessKwargs(ProcessCircuitsArgs, BaseModel):
    """Extra kwargs to process_circuits for Aer."""

    seed: Optional[int] = None


class AerSimulationRequest(BaseModel):
    """Aer config and any additional kwargs to process_circuits."""

    name: Literal["AerSimulationRequest"] = "AerSimulationRequest"
    circuits: list[CircuitModel]
    kwargs: AerProcessKwargs = AerProcessKwargs()
    config: AerConfig = AerConfig()



"""Aggregate models exposed at the top level."""

from typing import Literal, Union

from pydantic import BaseModel, Field, RootModel
from pytket.passes import BasePass

from quantinuum_schemas.aer import (
    AerCompilationRequest,
    AerConfig,
    AerSimulationRequest,
)
from quantinuum_schemas.qulacs import (
    QulacsCompilationRequest,
    QulacsConfig,
    QulacsSimulationRequest,
)

CompilationRequestData = AerCompilationRequest | QulacsCompilationRequest
SimulationRequestData = AerSimulationRequest | QulacsSimulationRequest
BackendConfig = AerConfig | QulacsConfig


class CompilationRequest(RootModel):
    """Root model for compilation requests."""

    root: CompilationRequestData = Field(..., discriminator="name")


class SimulationRequest(RootModel):
    """Root model for simulation requests."""

    root: SimulationRequestData = Field(..., discriminator="name")



"""Generael helper functions for Python typing."""
from typing import NoReturn


def assert_never(never: NoReturn) -> NoReturn:
    """Used to enforce exhaustiveness in a match statement.

    https://github.com/microsoft/pyright/issues/2569"""
    assert False, f"Unhandled type: f{never}"



"""Functions requiring extra imports."""

from pytket.backends.backend import Backend

from quantinuum_schemas import (
    AerConfig,
    BackendConfig,
    CompilationRequest,
    QulacsConfig,
    SimulationRequest,
)
from quantinuum_schemas.models import BackendResultModel, CircuitModel
from quantinuum_schemas.typing_utils import assert_never


# pylint: disable=import-outside-toplevel
def get_backend(config: BackendConfig) -> Backend:
    """Instantiate a backend from a config."""
    match config:
        case AerConfig():
            from pytket.extensions.qiskit.backends.aer import AerBackend

            return AerBackend(
                noise_model=None,
                simulation_method=config.simulation_method,
                crosstalk_params=None,
                n_qubits=config.n_qubits,
            )
        case QulacsConfig():
            from pytket.extensions.qulacs.backends.qulacs_backend import QulacsBackend

            return QulacsBackend(result_type=config.result_type)
        case _:
            assert_never(config)


def compile_circuit(req: CompilationRequest) -> CircuitModel:
    """Compile a circuit given a CompilationRequest."""

    backend = get_backend(req.root.config)
    base_pass = backend.default_compilation_pass(**req.root.kwargs.model_dump())
    circuit = req.root.circuit
    base_pass.apply(circuit)
    return circuit


def simulate_circuits(
    req: SimulationRequest,
) -> list[BackendResultModel]:
    """Simulate a circuit given a SimulationRequest."""

    backend = get_backend(req.root.config)
    return backend.run_circuits(req.root.circuits, **req.root.kwargs.model_dump())
